/**
 * Copyright (c) FOM-Nikhef 2015-
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2015-
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#define _XOPEN_SOURCE 500


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <fnmatch.h>
#include <ctype.h>

#include <lcmaps/lcmaps_log.h>

#include "parse_utils.h"


/************************************************************************
 * private prototypes
 ************************************************************************/

/* reads in next line from buffer starting at position myindex. Line is returned
 * in line and needs freeing.
 * \return 1 when line is found, 0 when end of buffer, -1 on error */
static int get_line(const char *buffer, size_t *myindex, char **line);

/* Matches against next entry in entries, starting at position myindex.
 * The recursion level is kept in recur_level.
 * \return 1 when a entry is found, 0 when no next entry, -1 on error, -2 on no
 * more entries */
static int match_next_entry(char *entries, size_t *myindex, int recur_level,
			    const char *infodir, const char *issuer_dn);

/* Tries to find issuer_dn among the issuer DNs (the value of the subjectdn key)
 * in the filename in infodir. The recursion level is passed in via recur_level.
 * \return 1 on match found, 0 on no match, -1 on error */
static int match_info_file(const char *infodir, const char *filename,
			   const char *subjectdn_key, const char *issuer_dn,
			   int recur_level);

/* url-decode input into output. output need freeing
 * \return 0 on success, -1 on error */
static int urldecode(const char *input, char **output);


/************************************************************************
 * public functions
 ************************************************************************/

/**
 * Reads in the given file into buffer. Buffer needs freeing.
 * return 0 on success, -1 on error
 */
int _vo_ca_ap_read_file(const char *filename, char **filebuffer)    {
    char *buffer=NULL;
    int fd=-1;
    struct stat stat_buf;
    size_t size;

    /* protect against invalid input */
    if (!filename || !filebuffer)   {
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* open file */
    if ( (fd=open(filename,O_RDONLY))<0)    {
	lcmaps_log(LOG_ERR, "%s: cannot open file %s: %s\n",
		       __func__, filename, strerror(errno));
	return -1;
    }

    /* get size */
    if (fstat(fd,&stat_buf)<0)    {
	lcmaps_log(LOG_ERR, "%s: cannot stat file %s: %s\n",
		   __func__, filename, strerror(errno));
	goto error;
    }

    /* allocate buffer: need one extra for \0 byte */
    size=(size_t)stat_buf.st_size;
    if ( (buffer=malloc(size+1))==NULL )    {
	lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
	goto error;
    }

    /* Only read in file when there actually is data */
    if (size>0)	{ /* read file */
	if (read(fd,buffer,size)!=(ssize_t)size) {
	    lcmaps_log(LOG_ERR, "%s: cannot read file %s: %s\n",
		       __func__, filename, strerror(errno));
	    goto error;
	}
    } else  { /* Probably not what you want... */
	lcmaps_log(LOG_INFO,"%s: file %s has size zero.\n",
		__func__, filename);
    }
    buffer[size]='\0';  /* Need to terminate, read() does not */

    /* all done succesfully */
    close(fd);
    *filebuffer=buffer;

    return 0;

error:
    if (fd>=0)
	close(fd);
    free(buffer);

    return -1;
}

/**
 * Search for given voname (or PSEUDO_VO when voname==NULL) in buffer. When
 * found, match entries against issuer_dn. For non-absolute `file:'-based
 * entries, the info file is looked for in infodir.
 * \return 1 when match found, 0 when no match found, -1 on error
 */
int _vo_ca_ap_match_vo_issuer(const char *buffer, const char *infodir,
			      const char *voname, const char *issuer_dn)
{
    char *line=NULL, *vopart=NULL, *remainder=NULL;
    int rc=-1;
    size_t i, pos=0, myindex;

    /* protect against invalid input */
    if (!buffer || !infodir || !issuer_dn)   {
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* Log what we are doing */
    lcmaps_log(LOG_DEBUG,
	    "%s: Looking for \"%s\" and issuer: \"%s\"\n",
	    __func__, voname ? voname : PSEUDO_VO, issuer_dn);

    /* Find line matching VO pattern */
    while ( (rc=get_line(buffer, &pos, &line))==1)	{
	/* Skip empty lines */
	if (line[0]=='\0')  {
	    free(line); line=NULL;
	    continue;
	}
	if (line[0]=='"')   {	/* look for end-double-quote */
	    for (i=1; line[i]!='"'; i++) {
		if (line[i]=='\0')  {
		    lcmaps_log(LOG_ERR,
			    "%s: cannot find end-quote on line %s\"",
			    __func__, line);
		    goto cleanup;
		}
	    }
	    vopart=line+1;
	} else { /* look for white-space */
	    for (i=0; line[i]!=' ' && line[i]!='\t'; i++) {
		if (line[i]=='\0')  {
		    lcmaps_log(LOG_ERR,
			    "%s: cannot find CA part on line %s\"",
			    __func__, line);
		    goto cleanup;
		}
	    }
	    vopart=line;
	}
	/* Truncate line */
	line[i]='\0';
	if (voname==NULL)   {	/* Non-VOMS: compare with - */
	    if (strcmp(vopart, PSEUDO_VO)==0) {
		remainder=line+i+1;
		break;
	    }
	} else	{   /* VOMS: do 'path' match */
	    if (fnmatch(vopart, voname, FNM_NOESCAPE)==0) {
		remainder=line+i+1;
		break;
	    }
	}
	/* Go to next line */
	free(line); line=NULL;
    }

    /* If we found a match, match using the entries part */
    if (remainder)	{
	/* Log that we have a hit */
	if (voname)
	    lcmaps_log(LOG_INFO, "%s: found VO: \"%s\"\n", __func__, voname);
	else
	    lcmaps_log(LOG_INFO, "%s: found entry: \"%s\"\n",
		    __func__, PSEUDO_VO);
	/* Now try to match the issuer DN entries */
	if (remainder[0]=='\0')	{
	    /* nothing left */
	    rc=-2;
	} else {
	    myindex=0;
	    /* continue till match (=1), eol (=-2) or error (-1).
	     * Start with recursion level 0 */
	    while ( (rc=match_next_entry(remainder, &myindex, 0,
					 infodir, issuer_dn))==0 );
	}
	goto cleanup;
    } else {
	/* Log that we failed */
	lcmaps_log(LOG_DEBUG, "%s: NO line matches: %s\n",
		__func__, voname ? voname : PSEUDO_VO);
    }


cleanup:
    free(line);

    /* Log when we ran out of entries */
    if (rc==-2)
	lcmaps_log(LOG_DEBUG, "%s: no entries left on line\n", __func__);

    return (rc==-2 ? 0 : rc);
}


/************************************************************************
 * Private functions
 ************************************************************************/

/**
 * reads in next line from buffer starting at position myindex. Line is returned
 * in line and needs freeing.
 * \return 1 when line is found, 0 when end of buffer, -1 on error
 */
static int get_line(const char *buffer, size_t *myindex, char **line)	{
    size_t pos1, pos2, len, i, j;
    char *tmpline=NULL;

    /* protect against invalid input */
    if (!buffer || !myindex || !line)	{
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* Move over leading whitespace and empty lines */
    for (pos1=*myindex;
	 buffer[pos1]==' ' || buffer[pos1]=='\t' || buffer[pos1]=='\n';
	 pos1++);

    /* Did we reach the end? */
    if (buffer[pos1]=='\0')
	return 0;

    /* Find end-of-line */
    pos2=pos1;
    while (buffer[pos2]!='\n' && buffer[pos2]!='\0')	{
	if (buffer[pos2]=='\\' && buffer[pos2+1]=='\n')	/* found cont char */
	    pos2+=2;
	else	/* skip next entry */
	    pos2++;
    }
    len=pos2-pos1;

    /* Comment line? */
    if (buffer[pos1]=='#')  {
	if ( (tmpline=strdup(""))==NULL )   {
	    lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
	    return -1;
	}
    } else {
	if ( (tmpline=malloc(len+1))==NULL )   {
	    lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
	    return -1;
	}
	/* copy input while removing continuation characters */
	for (i=pos1, j=0; i<pos2; i++)	{
	    if (buffer[i]=='\\' && buffer[i+1]=='\n')	{
		i++;
		continue;
	    }
	    tmpline[j++]=buffer[i];
	}
	tmpline[j]='\0';
    }

    /* Update parameters */
    *myindex=pos2;
    *line=tmpline;

    return 1;
}

/**
 * Matches against next entry in entries, starting at position myindex. Keep
 * track of recursion level to prevent infinite loops.
 * \return 1 when a entry is found, 0 when no next entry, -1 on error, -2 on no
 * more entries
 */
static int match_next_entry(char *entries, size_t *myindex, int recur_level,
			    const char *infodir, const char *issuer_dn)
{
    size_t pos1,pos2,pos3,pos4,pos5;
    char *issuer;
    int prefix, rc;

    /* Check empty remainder first */
    if (entries==NULL)
	return -2;

    /* protect against invalid input */
    if (!myindex || !infodir || !issuer_dn) {
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* Skip over leading white-space: pos1 will be start of the text */
    for (pos1=*myindex; entries[pos1]==' ' || entries[pos1]=='\t'; pos1++);

    /* Anything left? */
    if (entries[pos1]=='\0' || entries[pos1]=='#')
	return -2;

    /* Skip over leading text: pos2 will be on the : */
    for (pos2=pos1; entries[pos2]>='a' && entries[pos2]<='z'; pos2++);

    /* Do we have a prefix? */
    if (pos2==pos1) {
	/* no prefix */
	prefix=0;
	pos3=pos1;
    } else {
	/* Now we should have a : */
	prefix=1;
	if (entries[pos2]!=':')	{
	    lcmaps_log(LOG_ERR,
		"%s: Invalid entries: %s\n", __func__, entries+pos1);
	    return -1;
	}

	/* truncate at pos2 */
	entries[pos2]='\0';

	/* put pos3 on the first char after the : */
	pos3=pos2+1;
    }

    /* What's the first character of the actual field? */
    if (entries[pos3]=='"')	{ /* search end-quote */
	/* increase pos3 one more and find end quote */
	for (pos4=++pos3; entries[pos4]!='\0' && entries[pos4]!='"'; pos4++);
	if (entries[pos4]!='"')    {
	    lcmaps_log(LOG_ERR,
		    "%s: Missing end-quote in entries: %s\n",
		    __func__, entries+pos1);
	    return -1;
	}
	/* truncate entry (pos4 is end-quote) */
	entries[pos4]='\0';
	/* Find next comma or end-of-buffer */
	for (pos5=pos4+1;
	     entries[pos5]!=',' && entries[pos5]!='#' && entries[pos5]!='\0';
	     pos5++);
	/* pos5 is a comma a hash or a zero byte */
	if (entries[pos5]==',')    /* next entry */
	    pos5++;
	else	/* end (incl comment): truncate */
	    entries[pos5]='\0';
    } else { /* search end or comment, white-space or comma */
	for (pos4=pos3;
	     entries[pos4]!='\0' && entries[pos4]!='#' &&
	     entries[pos4]!=' ' && entries[pos4]!='\t' &&
	     entries[pos4]!=','; pos4++);
	if (entries[pos4]=='\0' || entries[pos4]=='#')    /* end */
	    pos5=pos4;
	else if (entries[pos4]==',')	/* next entry */
	    pos5=pos4+1;
	else {	/* white-space */
	    for (pos5=pos4+1;
		 entries[pos5]!=',' && entries[pos5]!='\0'; pos5++);
	    if (entries[pos5]==',')
		pos5++;
	}
	/* truncate entry (pos4 is first char after) */
	entries[pos4]='\0';
    }

    /* Log what we will do */
    lcmaps_log(LOG_DEBUG, "%s: trying %s (file level %d): \"%s\"\n",
	    __func__,
	    prefix ? entries+pos1 : "issuer", recur_level, 
	    entries+pos3);

    /* Now do matching: either match directly or use info file */
    if (prefix==0 || strcmp(entries+pos1, "issuer")==0) { /* direct */
	/* Do basic URL decoding... */
	if (urldecode(entries+pos3, &issuer)<0)  {
	    lcmaps_log(LOG_ERR, "%s: Cannot decode issuer %s\n",
		    __func__, entries+pos3);
	    return -1;
	}
	/* ...and match it */
	if (strcmp(issuer_dn, issuer)==0)   {
	    rc=1;
	    lcmaps_log(LOG_INFO, "%s: found issuer: \"%s\"\n",
		    __func__, issuer_dn);
	} else	{
	    rc=0;
	}
	free(issuer);
    } else if (strcmp(entries+pos1, "file")==0)    {
	/* Check we haven't reached the maximum recursion level */
	if (recur_level >= RECURSION_MAX)   {
	    lcmaps_log(LOG_ERR,
		    "%s: Too many recursive info files: %d\n",
		    __func__, recur_level);
	    return -1;
	}
	/* file: add one to the current file recursion level */
	rc=match_info_file(infodir, entries+pos3, "subjectdn",
			   issuer_dn, recur_level+1);
    } else {	/* error */
	lcmaps_log(LOG_ERR, "%s: Unknown prefix %s\n",
		__func__, entries+pos1);
	return -1;
    }

    /* Set index to next entry */
    *myindex=pos5;

    return rc;
}

/**
 * Tries to find issuer_dn among the issuer DNs (the value of the subjectdn key)
 * in the filename in infodir
 * \return 1 on match found, 0 on no match, -1 on error
 */
static int match_info_file(const char *infodir, const char *filename,
			   const char *subjectdn_key, const char *issuer_dn,
			   int recur_level)
{
    char *myfilename=NULL, *buffer=NULL, *line=NULL;
    const char *absfile;
    size_t len, linepos=0, pos1, pos2, pos3;
    int rc=-1;

    /* protect against invalid input */
    if (!infodir || !filename || !subjectdn_key || !issuer_dn)	{
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* Prefix with infodir if needed */
    if (filename[0]!='/')   {
	len=strlen(infodir)+strlen(filename)+2;
	if ( (myfilename=malloc(len)) == NULL) {
	    lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
	    return -1;
	}
	if (snprintf(myfilename, len, "%s/%s", infodir, filename)<0)   {
	    lcmaps_log(LOG_ERR,
		    "%s: Cannot prefix filename with %s: %s\n",
		    __func__, infodir, strerror(errno));
	    goto cleanup;
	}
	absfile=myfilename;
    } else  {
	absfile=filename;
    }

    /* Log what we are doing */
    lcmaps_log(LOG_DEBUG, "%s: reading file (file level %d): \"%s\"\n",
	    __func__, recur_level, absfile);

    /* Read in file into buffer */
    if (_vo_ca_ap_read_file(absfile, &buffer)!=0)
	goto cleanup;

    /* Find subjectdn key */
    while ( (rc=get_line(buffer, &linepos, &line))==1)	{
	/* Skip empty lines */
	if (line[0]=='\0')  {
	    free(line); line=NULL;
	    continue;
	}
	/* Find end of key */
	for (pos1=0;
	     line[pos1]!='\0' && line[pos1]!='#' &&
	     line[pos1]!='=' && line[pos1]!=' ' && line[pos1]!='\t';
	     pos1++);
	/* When we ended due to whitespace, find the end: should be = */
	if (line[pos1]==' ' || line[pos1]=='\t')    {
	    for(pos2=pos1+1; line[pos2]==' ' || line[pos2]=='\t'; pos2++);
	} else
	    pos2=pos1;
	/* */
	if (line[pos2]!='=')   {
	    lcmaps_log(LOG_ERR,
		    "%s: invalid key=value line \"%s\" in \"%s\"\n",
		    __func__, line, filename);
	    rc=-1;
	    goto cleanup;
	}
	/* Truncate line after key */
	line[pos1]='\0';
	if (strcmp(line, subjectdn_key)==0) {
	    /* Log that we found it */
	    lcmaps_log(LOG_DEBUG, "%s: found key \"%s\"\n",
		    __func__, subjectdn_key);
	    /* Found right key: check entries */
	    pos3=0;
	    /* continue till match (=1), eol (=-2) or error (-1). */
	    while ( (rc=match_next_entry(line+pos2+1, &pos3, recur_level,
					 infodir, issuer_dn))==0 );
	    goto cleanup;
	}
	/* Cleanup and move to next line */
	free(line); line=NULL;
    }

    /* Log that we don't have a subjectdn key */
    lcmaps_log(LOG_DEBUG, "%s: No key \"%s\" in file: \"%s\"\n",
	    __func__, subjectdn_key, filename);

cleanup:
    free(line);
    free(buffer);
    free(myfilename);

    /* Log when we ran out of entries */
    if (rc==-2)
	lcmaps_log(LOG_DEBUG, "%s: no entries left in file \"%s\"\n",
		__func__, filename);

    return (rc==-2 ? 0 : rc);
}

/**
 * url-decode input into output. output need freeing
 * \return 0 on success, -1 on error
 */
static int urldecode(const char *input, char **output)	{
    size_t pos, pos_out;
    int high,low;   /* Use int since all ctype functions use int */
    char *tmp;

    /* protect against invalid input */
    if (!input || !output)  {
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* One extra for '\0' */
    if ( (tmp=malloc(strlen(input)+1))==NULL )	{
	lcmaps_log(LOG_ERR, "%s: out of memory\n",__func__);
	return -1;
    }

    /* Scan over input */
    for (pos=pos_out=0; input[pos]; pos++)  {
	if (input[pos]=='%')	{
	    if (!isxdigit(input[pos+1]) || !isxdigit(input[pos+2])) {
		/* Not valid as hex */
		free(tmp);
		return -1;
	    }
	    /* Convert high digit */
	    if (isdigit(input[pos+1])) /* high <A */
		high=input[pos+1]-'0';
	    else    /* high >=A */
		high=10+(tolower(input[pos+1])-'a');
	    /* Convert low digit */
	    if (isdigit(input[pos+2]))	/* low <A */
		low=input[pos+2]-'0';
	    else    /* low >= A */
		low=10+(tolower(input[pos+2])-'a');
	    /* Copy into output: explicitly cast to signed char */
	    tmp[pos_out++]=(char)(high*16+low);
	    pos+=2;
	} else {
	    /* Copy into output */
	    tmp[pos_out++]=input[pos];
	}
    }

    /* truncate */
    tmp[pos_out]='\0';

    /* Set output */
    *output=tmp;

    return 0;
}
