/**
 * Copyright (c) FOM-Nikhef 2015-
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2015-
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#define _XOPEN_SOURCE 500

#include <string.h>
#include <stdio.h>
#include <errno.h>

#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>

#include <lcmaps/lcmaps_log.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
#include <lcmaps/lcmaps_types.h>


#include "plugin_utils.h"


/************************************************************************
 * private prototypes
 ************************************************************************/

/* Convert PEM string in certstring to chain of X509, returned in certstack.
 * Function is identical to lcmaps_pem_string_to_x509_chain()
 * return 0 on success, non-zero on error */
static int pem_string_to_x509_chain(STACK_OF(X509) **certstack,
				    char *certstring);

/* Obtains the issuer DN of the EEC
 * Return 0 on success, -1 on failure */
static int chain_to_issuer_dn(STACK_OF(X509) *chain, char **issuer_dn);


/************************************************************************
 * public functions
 ************************************************************************/

/**
 * Get issuer DN from LCMAPS framework
 * \return 0 on succes, -1 on error
 */
int _vo_ca_ap_get_issuer_dn(int argc, lcmaps_argument_t *argv,
			    char **issuer_dn)  {
    char *pem=NULL;
    void *value=NULL;
    STACK_OF(X509) *chain=NULL;
    int chain_needs_free=0;

    /* protect against invalid input */
    if (!issuer_dn) {
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* Try to get chain from LCMAPS */
    value=lcmaps_getArgValue("px509_chain", "STACK_OF(X509) *", argc, argv);
    if (value == NULL || (chain = *(STACK_OF(X509) **)value) == NULL ) {
	/* No valid chain found in LCMAPS framework, try to obtain from PEM
	 * string */
	lcmaps_log(LOG_DEBUG, "%s: no X.509 chain is set, trying pem string.\n",
		__func__);
	value=lcmaps_getArgValue("pem_string", "char *", argc, argv);
	if (value==NULL || (pem=*(char**)value) == NULL ) {
	    /* also not found: fatal error */
	    lcmaps_log(LOG_WARNING,
		    "%s: no chain or pemstring is set.\n", __func__);
	    return -1;
	}

	/* Convert pem string to chain */
	if (pem_string_to_x509_chain(&chain, pem)!=0)	{
	    lcmaps_log(LOG_WARNING,
		    "%s: cannot convert pemstring to chain.\n", __func__);
	    return -1;
	}
	
	/* Set flag to keep track of cleanup */
	chain_needs_free=1;
    }

    /* Get issuer DN from certificate chain */
    if (chain_to_issuer_dn(chain, issuer_dn)==-1)   {
	lcmaps_log(LOG_INFO, "%s: cannot get issuer DN from chain\n", __func__);
	return -1;
    }

    if (chain_needs_free)
	sk_X509_pop_free(chain, X509_free);

    return 0;
}

/**
 * Get VO list from from LCMAPS framework, either from list of FQANs or from
 * list of VOMS structures.
 * \return 0 on succes, -1 on error
 */
int _vo_ca_ap_get_vo_list(int argc, lcmaps_argument_t *argv,
			  char ***vo_list, int *nvos) {
    int i, j, nfqan=0, tmp_nvos=0, match=0;
    char **fqan_list=NULL, **tmp_vo_list=NULL;
    size_t len;
    char *pos;
    void *value;
    lcmaps_vomsdata_t *vomsdata=NULL;

    /* protect against invalid input */
    if (!vo_list || !nvos)  {
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* First get all VO FQANs from LCMAPS framework */
    fqan_list = (char**)getCredentialData(LCMAPS_VO_CRED_STRING, &nfqan);
    if (nfqan>0) {
	lcmaps_log(LOG_DEBUG, "%s: found %d FQAN(s) in credential data\n",
		__func__, nfqan);

	/* Now convert list of FQANs to list of VO names */
	if ( (tmp_vo_list=calloc((size_t)nfqan, sizeof(char*)))==NULL ) {
	    lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
	    return -1;
	}
	for (i=0; i<nfqan; i++)	{
	    len=strlen(fqan_list[i]);
	    /* Find second / in fqan when any */
	    if ( (pos=strchr(fqan_list[i], '/')) != NULL )
		len-=strlen(pos);
	    /* Check if we already have it */
	    match=0;
	    for (j=0; j<tmp_nvos; j++)  {
		if (strncmp(fqan_list[i],tmp_vo_list[j],len)==0)	{
		    match=1;
		    break;
		}
	    }
	    if (!match)	{
		if ( (tmp_vo_list[tmp_nvos]=malloc(len+1))==NULL )    {
		    lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
		    goto cleanup;
		}
		strncpy(tmp_vo_list[tmp_nvos], fqan_list[i], len);
		tmp_vo_list[tmp_nvos][len]='\0';
		tmp_nvos++;
	    }
	}
    } else {
	/* No FQANs registered, use the introspect/run arguments */
	lcmaps_log(LOG_DEBUG,
	    "%s: no FQANs registered by other plugins, "
	    "trying run/introspect args\n",
	    __func__);

	/* Get value of voms_data_list and try to obtain number of VOs contained
	 * in it */
	if ( ( value = lcmaps_getArgValue("voms_data_list",
					  "lcmaps_vomsdata_t *",
					  argc, argv) ) &&
	     (vomsdata = *(lcmaps_vomsdata_t **) value) != NULL &&
	     vomsdata->nvoms > 0)
	{
	    /* we have VOs */
	    tmp_nvos=vomsdata->nvoms;
	    if ( (tmp_vo_list=calloc((size_t)tmp_nvos, sizeof(char*)))==NULL ) {
		lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
		return -1;
	    }
	    /* Now extract the list of VO names */
	    for (i=0; i<tmp_nvos; i++)    {
		/* 2 extra: leading '/', trailing '\0' */
		len=strlen(vomsdata->voms[i].voname)+2;
		if ( (tmp_vo_list[i]=malloc(len))==NULL )    {
		    lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
		    goto cleanup;
		}
		tmp_vo_list[i][0]='/';
		strncpy(tmp_vo_list[i]+1, vomsdata->voms[i].voname, len-1);
		tmp_vo_list[i][len-1]='\0';
	    }
	}
    }

    lcmaps_log(LOG_DEBUG, "%s: found %d VOs\n", __func__, tmp_nvos);
    *nvos=tmp_nvos;
    *vo_list=tmp_vo_list;

    return 0;

cleanup:
    for (i=0; i<tmp_nvos; i++)
	free(tmp_vo_list[i]);
    free(tmp_vo_list);

    return -1;
}


/**
 * prefixes relfile with SECURITY_DIR into absfile, which needs freeing
 * \return 0 on success -1 on error
 */
int _vo_ca_ap_get_prefixed_file(const char *relfile, char **absfile)  {
    int len;
    char *path=NULL;

    /* protect against invalid input */
    if (!relfile || !absfile)	{
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* build-up secure prefix */
    len=snprintf(path, 0, SECURITY_DIR"%s", relfile);
    if (len<0)  {
	lcmaps_log(LOG_ERR,"%s: snprintf failed: %s\n",
	    __func__, strerror(errno));
	return -1;
    }
    /* one extra for trailing '\0' */
    if ( (path=malloc((size_t)(len+1)))==NULL )	{
	lcmaps_log(LOG_ERR,"%s: out of memory\n", __func__);
	return -1;
    }
    snprintf(path, (size_t)(len+1), SECURITY_DIR"%s", relfile);

    *absfile=path;

    return 0;
}


/************************************************************************
 * private functions
 ************************************************************************/

/**
 * Convert PEM string in certstring to chain of X509, returned in certstack.
 * Function is identical to lcmaps_pem_string_to_x509_chain()
 * return 0 on success, non-zero on error
 */
static int pem_string_to_x509_chain(STACK_OF(X509) **certstack,
				    char *certstring)
{
    STACK_OF(X509) *mystack=NULL;
    BIO *certbio = NULL;
    STACK_OF(X509_INFO) *sk=NULL;
    X509_INFO *xi;

    /* protect against invalid input */
    if (!certstack || !certstring)	{
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* initialize new X509 chain and memory BIO */
    if ( (mystack = sk_X509_new_null()) == NULL ||
	 (certbio = BIO_new_mem_buf(certstring, -1)) == NULL)
    {
	lcmaps_log(LOG_ERR,"%s: out of memory\n", __func__);
	return -1;
    }

    /* Convert PEM via memory BIO to stack of X509_INFO */
    sk=PEM_X509_INFO_read_bio(certbio, NULL, NULL, NULL);
    BIO_free(certbio);
    if (!sk)
	goto err;

    /* Loop over the stack and convert */
    while (sk_X509_INFO_num(sk)) {
	xi=sk_X509_INFO_shift(sk);
	if (xi->x509 != NULL) {
	    sk_X509_push(mystack, xi->x509);
	    xi->x509=NULL;
	}
	X509_INFO_free(xi);
    }
    sk_X509_INFO_free(sk);

    /* Check there is at least one certificate */
    if (!sk_X509_num(mystack))
	goto err;

    /* Cleanup and return success */
    *certstack=mystack;

    return 0;

err:
    /* Cleanup and error out */
    sk_X509_pop_free(mystack, X509_free);
    return 1;
}

/**
 * Obtains the issuer DN of the EEC
 * Return 0 on success, -1 on failure
 */
static int chain_to_issuer_dn(STACK_OF(X509) *chain, char **issuer_dn)	{
    X509 *cert;
    int i, depth, amount_of_CAs=0, eec_index;
    const int purpose=X509_PURPOSE_MIN + X509_PURPOSE_get_by_sname("sslclient");

    /* protect against invalid input */
    if (!chain || !issuer_dn)	{
	lcmaps_log(LOG_ERR,"%s: 1 or more input arguments is NULL.\n",__func__);
	return -1;
    }

    /* Go through chain to look for EEC */
    depth = sk_X509_num(chain);

    /* How many CA certs are there in the chain? */
    for (i = 0; i < depth; i++) {
	/* final argument to X509_check_purpose() is whether to check for CAness
	 * */
	if (X509_check_purpose(sk_X509_value(chain, i), purpose, 1))
	    amount_of_CAs++;
    }

    /* Getting the user cert (filtering the CAs and such) */
    eec_index = depth-(amount_of_CAs+1);

    /* Next check also catches depth==0 */
    if (eec_index < 0) {
	/* Counting the stack went wrong. Erroneous stack? */
	lcmaps_log(LOG_WARNING,
		"%s: cannot parse certificate chain: "
		"depth-(amount_of_CAs+1)=%d < 0\n", __func__, eec_index);
	return -1;
    }

    /* We got the EEC */
    cert=sk_X509_value(chain, eec_index);

    /* Get EEC DN, note: we cannot use X509_NAME_print_ex() since it cannot
     * print the / delimiter notation. */
    *issuer_dn=X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);

    return 0;
}
