/**
 * Copyright (c) FOM-Nikhef 2015-
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2015-
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#define _XOPEN_SOURCE 500

#include <string.h>
#include <stddef.h>
#include <syslog.h>
#include <stdlib.h>

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>

#if defined(HAVE_LCMAPS_LCMAPS_PLUGIN_PROTOTYPES_H)
#   include <lcmaps/lcmaps_plugin_prototypes.h>
#else
#   include "lcmaps_plugin_prototypes.h"
#endif

#include "plugin_utils.h"
#include "parse_utils.h"


/************************************************************************
 * defines
 ************************************************************************/

#define PLUGIN_PREFIX	"lcmaps_vo_ca_ap"

#define PLUGIN_RUN	0   /* full run mode */
#define PLUGIN_VERIFY	1   /* verify-only mode */


/************************************************************************
 * global variables
 ************************************************************************/

static char *vo_ca_ap_file=NULL;	/* Filename of VO-CA-AP file */
static char *certdir=NULL;	/* policy info directory (X509_CERT_DIR) */


/************************************************************************
 * private prototypes
 ************************************************************************/

/* called by plugin_run() and plugin_verify() */
static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv,
				int lcmaps_mode);


/************************************************************************
 * public functions
 ************************************************************************/

/**
 * Initialize function for plugin
 */
int plugin_initialize(int argc, char **argv) {
    const char * logstr = PLUGIN_PREFIX"-plugin_initialize()";
    char *value;
    int i;

    /* Log commandline parameters on debug */
    lcmaps_log(LOG_DEBUG,"%s: passed arguments:\n",logstr);
    for (i=0; i < argc; i++)
	lcmaps_log(LOG_DEBUG,"%s: arg %d is %s\n", logstr, i, argv[i]);

    /* Parse arguments, argv[0] = name of plugin, so start with i = 1 */
    for (i = 1; i < argc; i++) {
	/* check vo_ca_ap_file option (or similar variations) */
        if ( strcmp(argv[i], "-vo-ca-ap-file") == 0 ||
	     strcmp(argv[i], "-vo_ca_ap_file") == 0)
	{
	    /* check valid filename argument */
            if (argv[i + 1] == NULL || argv[i + 1][0]=='\0') {
		lcmaps_log(LOG_ERR,
		    "%s: option %s needs to be followed by valid filename\n",
		    logstr, argv[i]);
		goto fail_init_module;
	    }
	    /* free existing one and copy new one */
	    free(vo_ca_ap_file); vo_ca_ap_file=NULL;
	    if (argv[i+1][0]=='/')  { /* absolute path */
		if ( (vo_ca_ap_file = strdup(argv[i + 1])) == NULL)	{
		    lcmaps_log(LOG_ERR, "%s: out of memory\n", logstr);
		    goto fail_init_module;
		}
	    } else { /* relative path */
		if (_vo_ca_ap_get_prefixed_file(argv[i + 1],&vo_ca_ap_file)==-1)
		    goto fail_init_module;
	    }
	    /* log the name and increase arg counter */
	    lcmaps_log(LOG_DEBUG, "%s: Using VO-CA-AP file \"%s\".\n",
		    logstr, vo_ca_ap_file);
            i++;
        }
	/* check whether to disable wildcard matching */
	else if (strcmp(argv[i], "-certdir") == 0 ||
                 strcmp(argv[i], "-capath") == 0)
	{
	    /* check valid dirname argument */
            if (argv[i + 1] == NULL || argv[i + 1][0]=='\0') {
		lcmaps_log(LOG_ERR,
		    "%s: option %s needs to be followed by valid dirname\n",
		    logstr, argv[i]);
		goto fail_init_module;
	    }
	    /* free existing one and copy new one */
	    free(certdir); certdir=NULL;
	    if (argv[i+1][0]=='/')  { /* absolute path */
		if ( (certdir = strdup(argv[i + 1])) == NULL)	{
		    lcmaps_log(LOG_ERR, "%s: out of memory\n", logstr);
		    goto fail_init_module;
		}
	    } else { /* relative path */
		if (_vo_ca_ap_get_prefixed_file(argv[i + 1],&vo_ca_ap_file)==-1)
		    goto fail_init_module;
	    }
	    /* log the name and increase arg counter */
	    lcmaps_log(LOG_DEBUG, "%s: Using certdir \"%s\".\n",
		    logstr, certdir);
            i++;
        }
	/* any other argument is an error */
	else
	{
            lcmaps_log(LOG_ERR,
		    "%s: Unknown argument for plugin: %s (failure)\n",
		    logstr, argv[i]);
            goto fail_init_module;
        }
    }

    /* Check we have a VO-CA-AP file */
    if (vo_ca_ap_file==NULL || vo_ca_ap_file[0]=='\0')    {
        lcmaps_log(LOG_ERR,
	    "%s: Error: No VO-CA-AP file was provided. "
	    "Use the option -vo-ca-ap-file <file> to set one.\n", logstr);
        goto fail_init_module;
    }

    /* Set certdir from other means if not specified on the cmdline
     * including gLExec->LCMAPS */
    if (certdir==NULL)	{
	value=getenv("X509_CERT_DIR");
	certdir=value ? strdup(value) : strdup(DEFAULT_CERT_DIR);
	if (certdir==NULL)  {
	    lcmaps_log(LOG_ERR,"%s: Out of memory\n",logstr);
	    goto fail_init_module;
        }
	lcmaps_log(LOG_INFO,"%s: Using certificate CA Path: %s\n",
		logstr, certdir); 
    }

    return LCMAPS_MOD_SUCCESS;

fail_init_module:
    free(vo_ca_ap_file); vo_ca_ap_file = NULL;
    free(certdir); certdir = NULL;

    return LCMAPS_MOD_FAIL;
}

/**
 * Introspect function for plugin, defines which lcmaps framework data is made
 * available as commandline options for the plugin_run.
 */
int plugin_introspect(int *argc, lcmaps_argument_t **argv) {
    const char * logstr = PLUGIN_PREFIX"-plugin_introspect()";
    static lcmaps_argument_t argList[] = {
	{"px509_chain",   "STACK_OF(X509) *",    0,NULL},
	{"pem_string" ,   "char *",              0,NULL},
	{"voms_data_list","lcmaps_vomsdata_t *", 0,NULL}, 
	{NULL,            NULL,                 -1,NULL}
    };

    lcmaps_log(LOG_DEBUG,"%s: introspecting\n", logstr);

    *argv = argList;
    *argc = lcmaps_cntArgs(argList);
    lcmaps_log(LOG_DEBUG,"%s: address first argument: %p\n",
	    logstr, (void*)argList);

    return LCMAPS_MOD_SUCCESS;
}

/**
 * run function for plugin, wrapper around plugin_run_or_verify() with
 * PLUGIN_RUN mode
 */
int plugin_run(int argc, lcmaps_argument_t *argv) {
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}

/**
 * run function for plugin, wrapper around plugin_run_or_verify() with
 * PLUGIN_VERIFY mode
 */
int plugin_verify(int argc, lcmaps_argument_t *argv) {
    return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

/**
 * terminate plugin, frees used memory.
 * return LCMAPS_MOD_SUCCESS
 */
int plugin_terminate(void) {
    const char * logstr = PLUGIN_PREFIX"-plugin_terminate()";

    lcmaps_log(LOG_DEBUG,"%s: terminating\n", logstr);

    free(vo_ca_ap_file);
    vo_ca_ap_file=NULL;
    free(certdir);
    certdir=NULL;

    return LCMAPS_MOD_SUCCESS;
}


/************************************************************************
 * private functions
 ************************************************************************/

/**
 * actual plugin_run or plugin_verify function, which one depends on last
 * argument
 */
static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv,
                                int lcmaps_mode)
{
    const char *logstr=NULL;
    char *issuer_dn=NULL;
    char **vo=NULL;
    int i,nvoms=0;
    int mod_rc=LCMAPS_MOD_FAIL;
    char *buffer=NULL;

    /* Set suitable logstr */
    if (lcmaps_mode == PLUGIN_RUN)
        logstr = PLUGIN_PREFIX"-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = PLUGIN_PREFIX"-plugin_verify()";
    else {
        lcmaps_log(LOG_ERR, PLUGIN_PREFIX"-plugin_run_or_verify(): "
                "attempt to run plugin in invalid mode: %d\n", lcmaps_mode);
        goto cleanup;
    }

    /* get issuer DN from pem or chain in LCMAPS framework */
    if (_vo_ca_ap_get_issuer_dn(argc, argv, &issuer_dn)==-1)	{
	lcmaps_log(LOG_ERR, "%s: cannot get issuer DN from LCMAPS\n", logstr);
	goto cleanup;
    }

    /* get list of VO names from data in LCMAPS framework */
    if (_vo_ca_ap_get_vo_list(argc, argv, &vo, &nvoms)==-1)	{ 
	lcmaps_log(LOG_ERR, "%s: cannot get VO list from LCMAPS\n", logstr);
	goto cleanup;
    }
    
    /* Get VO-CA-AP file (uses get_file() ) */
    lcmaps_log(LOG_DEBUG,"%s: using VO-CA-AP file %s\n", logstr, vo_ca_ap_file);
    if (_vo_ca_ap_read_file(vo_ca_ap_file, &buffer)!=0)	{
	lcmaps_log(LOG_ERR, "%s: cannot read VO-CA-AP file %s\n",
	    logstr, vo_ca_ap_file);
	goto cleanup;
    }

    /* Now loop over voms and try to match each of them */
    if (nvoms>0)    {
	for (i=0; i<nvoms; i++) {
	    switch (_vo_ca_ap_match_vo_issuer(buffer,certdir,vo[i],issuer_dn)) {
		case -1:    /* error */
		    lcmaps_log(LOG_ERR,
			"%s: matching failed for VO \"%s\", issuer \"%s\"\n",
			logstr, vo[i], issuer_dn);
		    goto cleanup;
		case 0:	    /* no match */
		    lcmaps_log(LOG_NOTICE,
			"%s: No match for VO \"%s\", issuer \"%s\"\n",
			logstr, vo[i], issuer_dn);
		    goto cleanup;
		default:    /* match */
		    lcmaps_log(LOG_INFO,
			"%s: Found match for VO \"%s\", issuer \"%s\"\n",
			logstr, vo[i], issuer_dn);
	    }
	}
    } else {
	switch (_vo_ca_ap_match_vo_issuer(buffer,certdir,NULL,issuer_dn)) {
	    case -1:	/* error */
		lcmaps_log(LOG_ERR,
		    "%s: matching failed for issuer \"%s\"\n",
		    logstr, issuer_dn);
		goto cleanup;
	    case 0:	/* no match */
		lcmaps_log(LOG_NOTICE,
		    "%s: No match for issuer \"%s\"\n",
		    logstr, issuer_dn);
		goto cleanup;
	    default:
		lcmaps_log(LOG_INFO,
		    "%s: Found match for issuer \"%s\"\n",
		    logstr, issuer_dn);
	}
    }

    /* Plugin succeeded */
    mod_rc=LCMAPS_MOD_SUCCESS;

cleanup:
    if (vo) {
	for(i=0; i<nvoms; i++) free(vo[i]);
	free(vo);
    }
    free(buffer);
    free(issuer_dn);

    return mod_rc;
}
